from astroid import nodes  # type: ignore
from typing import TYPE_CHECKING, Optional, List, Union

from pylint.checkers import BaseChecker  # type: ignore

if TYPE_CHECKING:
    from pylint.lint import PyLinter  # type: ignore

from edulint.linting.checkers.utils import get_name, is_builtin, get_range_params
from edulint.linting.checkers.modified_listener import ModifiedListener


class ImproveForLoop(BaseChecker):  # type: ignore

    name = "improve-for-loop"
    msgs = {
        "R6101": (
            "Iterate directly: \"for var in %s\" (with appropriate name for \"var\")",
            "use-foreach",
            "Emitted when a for-range loop is used while a for-each loop would suffice.",
        ),
        "R6102": (
            "Iterate using enumerate: \"for %s, var in enumerate(%s)\" (with appropriate name for \"var\")",
            "use-enumerate",
            "Emitted when a for-range loop is used with the element at each index is accessed as well.",
        ),
    }

    def __init__(self, linter: Optional["PyLinter"] = None) -> None:
        super().__init__(linter)

    class StructureIndexedVisitor(ModifiedListener[bool]):
        default = False

        @staticmethod
        def combine(results: List[bool]) -> bool:
            return any(results)

        def __init__(self, structure: Union[nodes.Name, nodes.Attribute], index: nodes.Name):
            self.structure = structure
            self.index = index
            super().__init__([structure, index])

        def visit_subscript(self, subscript: nodes.Subscript) -> bool:
            sub_result = self.visit_many(subscript.get_children())
            if sub_result:
                return sub_result

            parent = subscript.parent
            if self.was_reassigned(self.structure, allow_definition=False) \
                    or self.was_reassigned(self.index, allow_definition=False):
                return False
            if not isinstance(subscript.value, type(self.structure)) \
                    or (isinstance(parent, nodes.Assign) and subscript in parent.targets) \
                    or not isinstance(subscript.slice, nodes.Name):
                return sub_result

            return subscript.slice.name == self.index.name and (
                (isinstance(self.structure, nodes.Name) and self.structure.name == subscript.value.name)
                or (isinstance(self.structure, nodes.Attribute) and self.structure.attrname == subscript.value.attrname)
            )

    class IndexUsedVisitor(ModifiedListener[bool]):
        default = False

        @staticmethod
        def combine(results: List[bool]) -> bool:
            return any(results)

        def __init__(self, structure: Union[nodes.Name, nodes.Attribute], index: nodes.Name):
            self.structure = structure
            self.index = index
            super().__init__([structure, index])

        def visit_name(self, name: nodes.Name) -> bool:
            super().visit_name(name)
            if name.name != self.index.name:
                return False
            if not isinstance(name.parent, nodes.Subscript) \
                    or not isinstance(self.structure, type(name.parent.value)) \
                    or self.was_reassigned(name, allow_definition=False):
                return True

            subscript = name.parent
            if not ((isinstance(subscript.value, nodes.Name)
                    and subscript.value.name == self.structure.name)
                    or (isinstance(subscript.value, nodes.Attribute)
                    and subscript.value.attrname == self.structure.attrname)):
                return True

            return isinstance(subscript.parent, nodes.Assign) and subscript in subscript.parent.targets

    def visit_for(self, node: nodes.For) -> None:
        range_params = get_range_params(node.iter)
        if range_params is None:
            return

        start, stop, step = range_params
        if not isinstance(start, nodes.Const) or start.value != 0 \
                or not isinstance(stop, nodes.Call) or not is_builtin(stop.func, "len") or len(stop.args) != 1 \
                or not isinstance(step, nodes.Const) or step.value != 1:
            return

        structure = stop.args[0]
        index = node.target
        if not isinstance(structure, nodes.Name) and not isinstance(structure, nodes.Attribute):
            return

        if self.StructureIndexedVisitor(structure, node.target).visit_many(node.body):
            structure_name = get_name(structure)
            if self.IndexUsedVisitor(structure, node.target).visit_many(node.body):
                self.add_message("use-enumerate", args=(index.name, structure_name), node=node)
            else:
                self.add_message("use-foreach", args=structure_name, node=node)


class NoGlobalVars(BaseChecker):
    name = "no-global-variables"
    msgs = {
        "R6401": (
            "Do not use global variables; you use %s, modifying it for example at line %i.",
            "no-global-vars",
            "Emitted when the code uses global variables."
        ),
    }

    def __init__(self, linter: "PyLinter"):
        super().__init__(linter)
        self.to_check = {}

    def visit_assignname(self, node: nodes.AssignName) -> None:
        frame = node.frame()
        if not isinstance(frame, nodes.Module):
            return

        if frame not in self.to_check:
            self.to_check[frame] = {}

        if node.name in self.to_check[frame].keys():
            return

        self.to_check[frame][node.name] = node

    def close(self) -> None:
        for frame, vars_ in self.to_check.items():
            listener = ModifiedListener(list(vars_.values()))
            listener.visit(frame)
            for node in vars_.values():
                if listener.was_modified(node, allow_definition=True):
                    nonglobal_modifiers = [n for n in listener.get_all_modifiers(node) if n.scope() != node.scope()]
                    if nonglobal_modifiers:
                        self.add_message("no-global-vars", node=node, args=(node.name, nonglobal_modifiers[0].lineno))


def register(linter: "PyLinter") -> None:
    """This required method auto registers the checker during initialization.
    :param linter: The linter to register the checker to.
    """
    linter.register_checker(ImproveForLoop(linter))
    linter.register_checker(NoGlobalVars(linter))



import pytest
from edulint.options import Option
from edulint.config.arg import Arg
from edulint.linting.problem import Problem
from utils import lazy_problem, apply_and_lint, create_apply_and_lint
from typing import List


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def xxx(x):",
        "    while True:",
        "        if x:",
        "            break",
    ], [
        lazy_problem().set_line(2)
        .set_text("The while condition can be replaced with '<negated x>'")
    ]),
])
def test_while_true_break(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(
        lines,
        [Arg(Option.PYLINT, "--disable=R1705"),
            Arg(Option.FLAKE8, "--extend-ignore=E501,F841")],
        [p.set_code("R6301") for p in expected_output]
    )


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "n = 10",
        "i = 0",
        "while i < n:",
        "    print('foo' + str(i))",
        "    i += 1",
    ], [
        lazy_problem().set_line(3)
        .set_text("Use for loop.")
    ]),
])
def test_use_for_loop_custom(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(
        lines,
        [Arg(Option.PYLINT, "--disable=all"),
            Arg(Option.PYLINT, "--enable=use-for-loop"),
            Arg(Option.NO_FLAKE8, "on")],
        [p.set_code("R6305") for p in expected_output]
    )


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "for i in range(10):",
        "    print('foo')",
        "    i += 1"
    ], [
        lazy_problem().set_line(3).set_text("Changing the control variable i of a for loop has no effect.")
    ]),
    ([
        "for _ in range(10):",
        "    _ = 'a,b,c,d'.split(',', 1)",
    ], [
    ]),
    ([
        "for _ in range(10):",
        "    _, rest = 'a,b,c,d'.split(',', 1)",
    ], [
    ]),
    ([
        "for _ in range(10):",
        "    for _ in range(10):"
        "        _, rest = 'a,b,c,d'.split(',', 1)",
    ], [
    ])
])
def test_changing_control_variable_custom(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(
        lines,
        [Arg(Option.PYLINT, "--disable=all"),
            Arg(Option.PYLINT, "--enable=changing-control-variable"),
            Arg(Option.NO_FLAKE8, "on")],
        [p.set_code("R6304") for p in expected_output]
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("hw14062.py", [
        lazy_problem().set_code("R6302").set_line(29)
        .set_text("Use tighter range boundaries, the first iteration never happens.")
    ]),
    ("hw34666.py", [
        lazy_problem().set_code("R6302").set_line(245)
        .set_text("Use tighter range boundaries, the last iteration never happens.")
    ]),
    ("m1630.py", [
        lazy_problem().set_code("R6302").set_line(107)
        .set_text("Use tighter range boundaries, the last iteration never happens.")
    ]),
])
def test_tighter_bounds_files(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--disable=all"), Arg(Option.PYLINT, "--enable=use-tighter-boundaries")],
        expected_output
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("010666-even.py", [lazy_problem().set_line(9)]),
    ("012630-next.py", []),
    ("012986-nested.py", [lazy_problem().set_line(35)]),
    ("013450-next.py", []),
    ("014286-coins.py", []),
    ("014556-prime.py", [lazy_problem().set_line(13).set_text("Use for loop.")]),
    ("014962-sequence.py", [lazy_problem().set_line(18).set_text("Use for loop.")]),
    ("015794-amicable.py", [lazy_problem().set_line(14)]),
    ("016119-coins.py", [lazy_problem().set_line(22)]),
    ("017667-coins.py", []),
    ("023140-credit.py", []),
    ("043611-swap_columns.py", []),
])
def test_use_for_loop_files(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--disable=all"), Arg(Option.PYLINT, "--enable=use-for-loop")],
        [problem.set_code("R6305") for problem in expected_output]
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("034636-right_cellular.py", []),
    ("104526-subset_sum.py", [
        lazy_problem().set_line(17)
        .set_text("Iterated structure nums is being modified inside the for loop body. "
                  "Use while loop or iterate over a copy."),
        lazy_problem().set_line(22)
    ]),
    ("035794-partition.py", []),
    ("hw24328.py", [
        lazy_problem().set_line(52)
        .set_text("Iterated structure row is being modified inside the for loop body. "
                  "Use while loop or iterate over a copy."),
        lazy_problem().set_line(53),
        lazy_problem().set_line(56),
        lazy_problem().set_line(99),
        lazy_problem().set_line(100),
        lazy_problem().set_line(103),
    ]),
    ("hw44081.py", [
        lazy_problem().set_line(149)
        .set_text("Iterated structure traders is being modified inside the for loop body. "
                  "Use while loop or iterate over a copy.")
    ]),
])
def test_modifying_iterated_files(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--disable=all"), Arg(Option.PYLINT, "--enable=modifying-iterated-structure")],
        [problem.set_code("R6303") for problem in expected_output]
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("hw33176.py", []),
    ("m4521.py", [
        lazy_problem().set_line(117).set_text("Inner for loop shadows outer for loop's control variable i.")
    ])
])
def test_shadowed_control_var_files(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--disable=all"), Arg(Option.PYLINT, "--enable=loop-shadows-control-variable"),
            Arg(Option.NO_FLAKE8, "on")],
        expected_output
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("014659-next.py", [
        lazy_problem().set_line(24).set_text("Changing the control variable i of a for loop has no effect.")
    ]),
    ("hw24226.py", [
        lazy_problem().set_line(77).set_text("Changing the control variable j of a for loop has no effect."),
        lazy_problem().set_line(101).set_text("Changing the control variable p of a for loop has no effect."),
    ]),
    ("hw24387.py", [
        lazy_problem().set_line(98).set_text("Changing the control variable i of a for loop has no effect."),
        lazy_problem().set_line(129),
    ]),
    ("hw34406.py", []),
    ("hw34666.py", [
        lazy_problem().set_line(40),
        lazy_problem().set_line(43),
        lazy_problem().set_line(51),
        lazy_problem().set_line(54),
        lazy_problem().set_line(57),
        lazy_problem().set_line(71),
        lazy_problem().set_line(74),
        lazy_problem().set_line(77),
    ]),
])
def test_changing_control_var_files(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--disable=all"), Arg(Option.PYLINT, "--enable=changing-control-variable")],
        [problem.set_code("R6304") for problem in expected_output]
    )



from astroid import nodes  # type: ignore
from typing import TYPE_CHECKING, Optional, Tuple

from pylint.checkers import BaseChecker  # type: ignore
from pylint.checkers.utils import only_required_for_messages

if TYPE_CHECKING:
    from pylint.lint import PyLinter  # type: ignore

from edulint.linting.checkers.utils import get_name, get_assigned_to, is_any_assign


class SimplifiableIf(BaseChecker):  # type: ignore

    name = "simplifiable-if"
    msgs = {
        "R6201": (
            "The if statement can be replaced with 'return %s'",
            "simplifiable-if-return",
            "Emitted when the condition of an if statement can be returned directly (possibly negated).",
        ),
        "R6202": (
            "The if statement can be replaced with 'return %s'",
            "simplifiable-if-return-conj",
            "Emitted when the condition of an if statement and the returned values "
            "can be combined using logical operators."
        ),
        "R6203": (
            "The conditional assignment can be replace with '%s = %s'",
            "simplifiable-if-assignment",
            "Emitted when the condition of an if statement can be assigned directly (possibly negated)."
        ),
        "R6210": (
            "The conditional assignment can be replace with '%s = %s'",
            "simplifiable-if-assignment-conj",
            "Emitted when the condition of an if statement and the assigned values "
            "can be combined using logical operators."
        ),
        "R6204": (
            "The if expression can be replaced with '%s'",
            "simplifiable-if-expr",
            "Emitted when the condition of an if expression can be returned directly (possibly negated)."
        ),
        "R6209": (
            "The if expression can be replaced with '%s'",
            "simplifiable-if-expr-conj",
            "Emitted when the condition of an if expression and the returned values "
            "can be combined using logical operators."
        ),
        "R6205": (
            "Use 'if %s: <else body>' instead of 'pass'",
            "simplifiable-if-pass",
            "Emitted when there is an if condition with a pass in the positive branch."
        ),
        "R6206": (
            "Both branches should return a value explicitly (one returns implicit None)",
            "no-value-in-one-branch-return",
            "Emitted when one branch returns a value and the other just returns."
        ),
        "R6207": (
            "The if statement can be merged with the nested one to 'if %s:'",
            "simplifiable-if-nested",
            "Emitted when the condition of an if statement can be merged "
            "with its nested if's condition using logical operators."
        ),
        "R6208": (
            "The if statement can be merged with the following one to 'if %s:'",
            "simplifiable-if-seq",
            "Emitted when the condition of an if statement can be merged "
            "with the next if's condition using logical operators."
        ),
    }

    def _is_bool(self, node: nodes.NodeNG) -> bool:
        return isinstance(node, nodes.Const) and node.pytype() == "builtins.bool"

    def _simplifiable_if_message(self, node: nodes.If, then: nodes.NodeNG, new_cond: str, only_replaces: bool) -> None:
        extra = "" if only_replaces else "-conj"
        if isinstance(node, nodes.IfExp):
            self.add_message("simplifiable-if-expr" + extra, node=node, args=(new_cond))
        elif isinstance(then, nodes.Return):
            self.add_message("simplifiable-if-return" + extra, node=node, args=(new_cond))
        else:
            self.add_message("simplifiable-if-assignment" + extra, node=node,
                             args=(get_name(get_assigned_to(then)[0]), new_cond))

    def _get_refactored(self, *args) -> str:
        result = []
        i = 0
        while i < len(args):
            arg = args[i]
            if isinstance(arg, str) and arg == "not":
                refactored = args[i + 1].as_string()
                if isinstance(args[i+1], nodes.BoolOp):
                    result.append(f"<negated ({refactored})>")
                else:
                    result.append(f"<negated {refactored}>")
                i += 2
            elif isinstance(arg, str) and arg in ("and", "or"):
                prev = args[i - 1]
                next_ = args[i + 1]
                if isinstance(prev, nodes.BoolOp) and prev.op != arg and (i - 2 < 0 or args[i - 2] != "not"):
                    result[-1] = f"({result[-1]})"
                result.append(arg)
                i += 1
                if isinstance(next_, nodes.BoolOp) and next_.op != arg:
                    result.append(f"({next_.as_string()})")
                    i += 1
            elif isinstance(arg, nodes.NodeNG):
                result.append(arg.as_string())
                i += 1

        return " ".join(result)

    def _names_assigned_to(self, node: nodes.NodeNG):
        return sorted([get_name(t) for t in get_assigned_to(node)])

    def _get_then_orelse(self, node: nodes.If) -> Optional[Tuple[nodes.NodeNG, nodes.NodeNG]]:
        if len(node.body) != 1 or len(node.orelse) > 1:
            return None

        then = node.body[0]

        if len(node.orelse) == 1:
            return then, node.orelse[0]

        after_if = node.next_sibling()
        if not isinstance(then, nodes.Return) or not isinstance(after_if, nodes.Return):
            return None

        return then, after_if

    def _refactored_cond_from_then_orelse(self, node: nodes.If, then_value: nodes.NodeNG, orelse_value: nodes.NodeNG) \
            -> Optional[Tuple[str, bool]]:
        then_bool_value = then_value.value if self._is_bool(then_value) else None
        orelse_bool_value = orelse_value.value if self._is_bool(orelse_value) else None

        if then_bool_value is None and orelse_bool_value is None:
            return None

        if then_bool_value is not None and orelse_bool_value is not None:
            if then_bool_value and not orelse_bool_value:
                return self._get_refactored(node.test), True
            if not then_bool_value and orelse_bool_value:
                return self._get_refactored("not", node.test), True
            return None

        if then_bool_value is not None and then_bool_value:
            return self._get_refactored(node.test, "or", orelse_value), False

        if then_bool_value is not None and not then_bool_value:
            return self._get_refactored("not", node.test, "and", orelse_value), False

        if orelse_bool_value is not None and orelse_bool_value:
            return self._get_refactored("not", node.test, "or", then_value), False

        if orelse_bool_value is not None and not orelse_bool_value:
            return self._get_refactored(node.test, "and", then_value), False

        assert False, "unreachable"

    def _merge_nested(self, node: nodes.If) -> None:
        refactored = self._get_refactored(node.test, "and", node.body[0].test)
        self.add_message("simplifiable-if-nested", node=node, args=refactored)

    def _same_values(self, node1: nodes.NodeNG, node2: nodes.NodeNG) -> bool:
        return (node1 is None and node2 is None) \
             or (node1 is not None and node2 is not None and node1.as_string() == node2.as_string())

    def _mergeable(self, node1: nodes.NodeNG, node2: nodes.NodeNG) -> bool:
        if not isinstance(node1, type(node2)):
            return False

        if isinstance(node1, nodes.Return):
            return node1.value is not None and node2.value is not None

        if is_any_assign(node1) and not isinstance(node1, nodes.For):
            assigned_to1 = get_assigned_to(node1)
            assigned_to2 = get_assigned_to(node2)
            return len(assigned_to1) == 1 and len(assigned_to1) == len(assigned_to2) \
                and isinstance(assigned_to1[0], type(assigned_to2[0])) \
                and not isinstance(assigned_to1[0], nodes.Subscript) \
                and self._names_assigned_to(node1) == self._names_assigned_to(node2)

        return False

    def _merge_sequential(self, node: nodes.If) -> None:
        second_if = node.next_sibling()

        if isinstance(node.test, nodes.BoolOp) or isinstance(second_if.test, nodes.BoolOp) \
                or len(second_if.orelse) != 0 or len(node.body) != 1 or len(second_if.body) != 1:
            return

        body1 = node.body[0]
        body2 = second_if.body[0]
        if not isinstance(body1, type(body2)) \
                or not isinstance(body1, nodes.Return) \
                or not self._same_values(body1.value, body2.value):
            return

        parent_body = node.parent.body
        if all(isinstance(n, nodes.If) or isinstance(n, nodes.Return)
                or (is_any_assign(n) and not isinstance(n, nodes.For)) for n in parent_body) \
                and sum(1 if isinstance(n, nodes.If) else 0 for n in parent_body) > 2:
            return

        refactored = self._get_refactored(node.test, "or", second_if.test)
        self.add_message("simplifiable-if-seq", node=node, args=(refactored))

    def _is_just_returning_if(self, node: Optional[nodes.NodeNG]) -> bool:
        return node is not None and isinstance(node, nodes.If) and isinstance(node.body[-1], nodes.Return)

    @only_required_for_messages("simplifiable-if-return", "simplifiable-if-return-conj",
                                "simplifiable-if-assignment", "simplifiable-if-assignment-conj",
                                "simplifiable-if-pass", "no-value-in-one-branch-return",
                                "simplifiable-if-nested", "simplifiable-if-seq")
    def visit_if(self, node: nodes.If) -> None:
        if len(node.orelse) == 0:
            if len(node.body) == 1 and isinstance(node.body[0], nodes.If) and len(node.body[0].orelse) == 0:
                self._merge_nested(node)
                return

            if node.next_sibling() is not None and isinstance(node.next_sibling(), nodes.If):
                self._merge_sequential(node)
                return

        if len(node.body) == 1 and isinstance(node.body[0], nodes.Pass) and len(node.orelse) > 0:
            self.add_message("simplifiable-if-pass", node=node, args=(self._get_refactored("not", node.test)))
            return

        then_orelse = self._get_then_orelse(node)

        if not then_orelse:
            return

        then, orelse = then_orelse

        if isinstance(then, nodes.Return) and isinstance(orelse, nodes.Return) \
                and (then.value is None) != (orelse.value is None):
            self.add_message("no-value-in-one-branch-return", node=node)
            return

        if not self._mergeable(then, orelse):
            return

        if len(node.orelse) == 0 and self._is_just_returning_if(node.previous_sibling()):
            return

        refactored = self._refactored_cond_from_then_orelse(node, then.value, orelse.value)

        if refactored is not None:
            new_cond, only_replaces = refactored
            self._simplifiable_if_message(node, then, new_cond, only_replaces)

    @only_required_for_messages("simplifiable-if-expr", "simplifiable-if-expr-conj")
    def visit_ifexp(self, node: nodes.IfExp) -> None:
        then, orelse = node.body, node.orelse
        assert then is not None and orelse is not None

        if (not isinstance(then, nodes.Const) and not isinstance(then, nodes.Name)) or \
                (not isinstance(orelse, nodes.Const) and not isinstance(orelse, nodes.Name)):
            return

        refactored = self._refactored_cond_from_then_orelse(node, then, orelse)
        if refactored is not None:
            new_cond, only_replaces = refactored
            self._simplifiable_if_message(node, then, new_cond, only_replaces)


def register(linter: "PyLinter") -> None:
    """This required method auto registers the checker during initialization.
    :param linter: The linter to register the checker to.
    """
    linter.register_checker(SimplifiableIf(linter))



from astroid import nodes  # type: ignore
from typing import Optional, List, Union, TypeVar, Iterator, Dict
import re

from edulint.linting.checkers.utils import BaseVisitor, Named, get_name


T = TypeVar("T")


class ModifiedListener(BaseVisitor[T]):

    NON_PURE_METHODS = re.compile(r"append|clear|extend|insert|pop|remove|reverse|sort|add|.*update|write")

    def __init__(self, watched: List[Named]):
        self.watched = watched
        self.modified = {get_name(var): [] for var in watched}
        self.stack = [{get_name(var): var.scope() for var in watched}]
        super().__init__()

    def _init_var_in_scope(self, node: nodes.NodeNG) -> T:
        self.stack[-1][get_name(node)] = node.scope()

    def visit_functiondef(self, node: nodes.FunctionDef) -> T:
        self.stack.append({})
        for arg in node.args.args:
            self._init_var_in_scope(arg)

        result = self.visit_many(node.get_children())

        self.stack.pop()
        return result

    def visit_global(self, node: nodes.Global) -> T:
        for name in node.names:
            self.stack[-1][name] = node.root().scope()

        return self.default

    def visit_nonlocal(self, node: nodes.Nonlocal) -> T:
        for name in node.names:
            self.stack[-1][name] = node.scope().parent.scope()

        return self.default

    def was_modified(self, node: nodes.NodeNG, allow_definition: bool) -> bool:
        return len(self.get_all_modifiers(node)) > (1 if allow_definition else 0)

    @staticmethod
    def _reassigns(node: nodes.NodeNG) -> bool:
        return type(node) in (nodes.AssignName, nodes.AssignAttr, nodes.DelName, nodes.DelAttr)

    def was_reassigned(self, node: nodes.NodeNG, allow_definition: bool) -> bool:
        return sum(self._reassigns(mod) for mod in self.get_all_modifiers(node)) > (1 if allow_definition else 0)

    def get_all_modifiers(self, node: nodes.NodeNG) -> List[nodes.NodeNG]:
        return self.modified[get_name(node)]

    def get_sure_modifiers(self, node: nodes.NodeNG) -> List[nodes.NodeNG]:
        result = []
        for modifier in self.get_all_modifiers(node):
            if ModifiedListener._reassigns(modifier):
                break
            result.append(modifier)
        return result

    @staticmethod
    def _strip(node: nodes.NodeNG) -> Optional[Union[nodes.Name, nodes.AssignName]]:
        while True:
            if isinstance(node, nodes.Subscript):
                node = node.value
            elif type(node) in (nodes.Attribute, nodes.AssignAttr, nodes.DelAttr):
                node = node.expr
            else:
                break

        return node if type(node) in (nodes.Name, nodes.AssignName, nodes.DelName) else None

    def _get_var_scope(self, var: str) -> Optional[Dict[str, nodes.NodeNG]]:
        for scope in reversed(self.stack):
            if var in scope:
                return scope
        return None

    def _is_same_var(self, var: Named, node: Named) -> bool:
        varname = get_name(var)
        return varname == get_name(node) and self._get_var_scope(varname)[varname] == var.scope()

    def _visit_assigned_to(self, node: nodes.NodeNG) -> None:
        stripped = self._strip(node)
        if stripped is None:
            return

        if isinstance(node, nodes.AssignName) and get_name(stripped) not in self.stack[-1]:
            self._init_var_in_scope(stripped)

        for var in self.watched:
            if self._is_same_var(var, stripped):
                self.modified[get_name(var)].append(node)

    def _names_from_tuple(self, targets: List[nodes.NodeNG]) -> Iterator[nodes.NodeNG]:
        for target in targets:
            if not isinstance(target, nodes.Tuple):
                yield target
            else:
                yield from self._names_from_tuple(target.elts)

    def visit_assign(self, assign: nodes.Assign) -> T:
        for target in self._names_from_tuple(assign.targets):
            self._visit_assigned_to(target)

        return self.visit_many(assign.get_children())

    def visit_augassign(self, node: nodes.AugAssign) -> T:
        self._visit_assigned_to(node.target)
        return self.visit_many(node.get_children())

    def visit_annassign(self, node: nodes.AnnAssign) -> T:
        self._visit_assigned_to(node.target)
        return self.visit_many(node.get_children())

    def visit_attribute(self, node: nodes.Attribute) -> T:
        if isinstance(node.parent, nodes.Call) \
                and ModifiedListener.NON_PURE_METHODS.match(node.attrname):

            stripped = self._strip(node)
            if stripped is None:
                return self.visit_many(node.get_children())

            for var in self.watched:
                if self._is_same_var(var, stripped):
                    self.modified[get_name(var)].append(node.parent)

        return self.visit_many(node.get_children())

    def visit_del(self, node: nodes.Delete) -> T:
        for target in node.targets:
            if isinstance(target, nodes.DelName):
                scope = self._get_var_scope(target.name)
                if scope is not None:
                    scope.pop(target.name)

            stripped = self._strip(target)

            for var in self.watched:
                if self._is_same_var(var, stripped):
                    self.modified[get_name(var)].append(
                        target if type(target) in (nodes.DelName, nodes.DelAttr) else node
                    )

        return self.visit_many(node.get_children())

    def visit_for(self, node: nodes.For) -> T:
        self._visit_assigned_to(node.target)
        return self.visit_many(node.body)



from edulint.linting.checkers.basic_checker import ModifiedListener
from edulint.linting.checkers.utils import get_name

from typing import List, Tuple, Dict
import astroid  # type: ignore
from astroid import nodes as astns
import pytest


def split_n_lines(nodes: List[astns.NodeNG], lines: int) -> Tuple[List[astns.NodeNG], List[astns.NodeNG]]:
    if lines == 0:
        return [], nodes

    consumed_lines = []
    for i, node in enumerate(nodes):
        consumed_lines.append(node)
        if hasattr(node, "body"):
            consumed_lines_rec, result_rec = split_n_lines(node.body, lines - len(consumed_lines))
            consumed_lines += consumed_lines_rec
        else:
            result_rec = []

        if len(consumed_lines) == lines:
            return consumed_lines, result_rec + nodes[i+1:]
    return consumed_lines, []


@pytest.mark.filterwarnings("ignore:The 'Module.doc' attribute")
@pytest.mark.parametrize("program,init_lines,before_types,after_types", [
    ("x = 5", 1, [astns.Assign], []),
    ("""
x = 5
x += 1""", 1, [astns.Assign], [astns.AugAssign]),
    ("""
x = 5
x = 1""", 1, [astns.Assign], [astns.Assign]),
    ("""
x = 5
y = 1
x += 1""", 2, [astns.Assign, astns.Assign], [astns.AugAssign]),
    ("""
for x in range(10):
    x = 1""", 1, [astns.For], [astns.Assign]),
    ("""
for x in range(10):
    for y in range(11):
        x = 1""", 1, [astns.For], [astns.For]),
    ("""
for x in range(10):
    for y in range(11):
        x = 1""", 2, [astns.For, astns.For], [astns.Assign]),
    ("""
for x in range(10):
    x = 1
    for y in range(11):
        x = 1""", 2, [astns.For, astns.Assign], [astns.For]),
    ("""
for x in range(10):
    x = 1
    for y in range(11):
        x = 1""", 1, [astns.For], [astns.Assign, astns.For]),
    ("""
for x in range(10):
    x = 1
    for y in range(11):
        x = 1
    x += 1""", 1, [astns.For], [astns.Assign, astns.For, astns.AugAssign]),
    ("""
for x in range(10):
    x = 1
    for y in range(11):
        x = 1
    x += 1""", 2, [astns.For, astns.Assign], [astns.For, astns.AugAssign]),
    ("""
for x in range(10):
    x = 1
    for y in range(11):
        x = 1
    x += 1""", 3, [astns.For, astns.Assign, astns.For], [astns.Assign, astns.AugAssign]),
    ("""
for x in range(10):
    x = 1
    for y in range(11):
        x = 1
    x += 1""", 4, [astns.For, astns.Assign, astns.For, astns.Assign], [astns.AugAssign]),
])
def test_split_n_lines(program: str, init_lines: int, before_types: List[type], after_types: List[type]):
    module = astroid.parse(program)
    print(module)
    before, after = split_n_lines(module.body, init_lines)

    def compare(nodes: List[astns.NodeNG], node_types: List[type]):
        assert len(nodes) == len(node_types)

        for node, node_type in zip(nodes, node_types):
            assert isinstance(node, node_type)

    compare(before, before_types)
    compare(after, after_types)


@pytest.mark.parametrize("program,init_lines,modified", [
    (["x = 5"], 1, {"x": False}),
    ([
        "x = 5",
        "x += 1"
    ], 1, {"x": True}),
    ([
        "x = 5",
        "x = 1"
    ], 1, {"x": True}),
    ([
        "x = 5",
        "y = 1",
        "x += 1"
    ], 2, {"x": True, "y": False}),
    ([
        "for x in range(10):",
        "    for x in range(11):",
        "        pass"
    ], 1, {"x": True}),
    ([
        "for x in range(10):",
        "    for y in range(11):",
        "        pass"
    ], 1, {"x": False}),
    ([
        "x = 0",
        "def foo():",
        "    global x",
        "    x = 1"
    ], 1, {"x": True}),
    ([
        "x = 0",
        "def foo():",
        "    global x",
        "    x += 1"
    ], 1, {"x": True}),
    ([
        "x = 0",
        "def foo():",
        "    nonlocal x",
        "    x = 1"
    ], 1, {"x": True}),
    ([
        "x = 0",
        "def foo():",
        "    x = 1",
        "    def bar():",
        "        nonlocal x",
        "        x = 2"
    ], 1, {"x": False}),
    ([
        "x = []",
        "def foo():",
        "    x.append(1)"
    ], 1, {"x": True}),
    ([
        "x = []",
        "def foo():",
        "    x[0] = 1"
    ], 1, {"x": True}),
    ([
        "x = None",
        "def foo():",
        "    x.y = 1"
    ], 1, {"x": True}),
    ([
        "x = None",
        "def foo():",
        "    x.y.z = 1"
    ], 1, {"x": True}),
    ([
        "x = None",
        "def foo():",
        "    x[0].y[1].z = 1"
    ], 1, {"x": True}),
    ([
        "x = None",
        "def foo():",
        "    x = []",
        "    x.append(0)"
    ], 1, {"x": False}),
    ([
        "x = None",
        "def foo():",
        "    x, y = [], 0",
        "    x.append(0)"
    ], 1, {"x": False}),
    ([
        "x = None",
        "def foo():",
        "    (x, y) = [], 0",
        "    x.append(0)"
    ], 1, {"x": False}),
    ([
        "x = None",
        "def foo(x):",
        "    x.append(0)"
    ], 1, {"x": False}),
    ([
        "x = None",
        "def foo(x: List[int] = []):",
        "    x.append(0)"
    ], 1, {"x": False}),
])
def test_modified_listener(program: List[str], init_lines: int, modified: Dict[str, bool]):
    module = astroid.parse("\n".join(program))
    before, after = split_n_lines(module.body, init_lines)

    def extract_vars(nodes: List[astns.NodeNG]) -> List[astns.NodeNG]:
        result = []
        for node in nodes:
            if hasattr(node, "target"):
                result.append(node.target)
            if hasattr(node, "targets"):
                result.extend(node.targets)
        return result

    watched = extract_vars(before)
    listener: ModifiedListener = ModifiedListener(watched)
    listener.visit_many(after)
    assert {get_name(n): listener.was_modified(n, allow_definition=False) for n in watched} == modified



import pytest
from edulint.options import Option
from edulint.config.arg import Arg
from edulint.linting.problem import Problem
from utils import lazy_problem, apply_and_lint, create_apply_and_lint
from typing import List


class TestShort:
    @pytest.mark.parametrize("lines,expected_output", [
        ([
            "if x > y:",
            "    print('foo')",
            "elif x <= y:",
            "    print('bar')"
        ], [
            lazy_problem().set_line(3).set_code("R6611")
        ]),
        ([
            "if x > y:",
            "    print('foo')",
            "if x <= y:",
            "    print('bar')"
        ], [
            lazy_problem().set_line(3).set_code("R6611")
        ]),
        ([
            "if x > y:",
            "    print('foo')",
            "if x <= y:",
            "    print('bar')",
            "else:",
            "    print('foobar')"
        ], [
        ]),
    ])
    def test_short_custom(self, lines: List[str], expected_output: List[Problem]) -> None:
        create_apply_and_lint(
            lines,
            [Arg(Option.PYLINT, "--disable=all"),
             Arg(Option.PYLINT, "--enable=short-problems"),
             Arg(Option.NO_FLAKE8, "on")],
            expected_output
        )

    @pytest.mark.parametrize("lines,expected_output", [
        ([
            "def foo(a):",
            "    a = a + 1",
        ], [lazy_problem()]),
        ([
            "def foo(a):",
            "    a = a + [0]",
        ], []),
        ([
            "def foo(a):",
            "    a = [0] + a",
        ], []),
        ([
            "def foo(a):",
            "    a = a + 'a'",
        ], [lazy_problem()]),
        ([
            "def foo(a):",
            "    a = 'a' + a",
        ], []),
        ([
            "def foo(a):",
            "    a = a & 2",
        ], [lazy_problem()]),
        ([
            "def foo(a):",
            "    a = a & {'a'}",
        ], []),
    ])
    def test_augassign_custom(self, lines: List[str], expected_output: List[Problem]) -> None:
        create_apply_and_lint(
            lines,
            [Arg(Option.PYLINT, "--disable=all"),
             Arg(Option.PYLINT, "--enable=use-augmenting-assignment"),
             Arg(Option.NO_FLAKE8, "on")],
            [problem.set_code("R6609") for problem in expected_output]
        )

    @pytest.mark.parametrize("lines,expected_output", [
        (["1 is True"], [lazy_problem()]),
        (["1 is not True"], [lazy_problem()]),
        (["1 is None"], []),
        (["1 is not None"], []),
    ])
    def test_no_is_custom(self, lines: List[str], expected_output: List[Problem]) -> None:
        create_apply_and_lint(
            lines,
            [Arg(Option.PYLINT, "--disable=all"),
             Arg(Option.PYLINT, "--enable=no-is-bool"),
             Arg(Option.NO_FLAKE8, "on")],
            [problem.set_code("R6613") for problem in expected_output]
        )

    @pytest.mark.parametrize("filename,expected_output", [
        ("044050-vigenere.py", [
            lazy_problem().set_code("R6614").set_line(38)
            .set_text("Use \"ord('A')\" instead of using the magical constant 65."),
            lazy_problem().set_code("R6615").set_line(39)
            .set_text("Remove the call to 'ord' and compare to the string directly \"> 'Z'\" instead of using "
                      "the magical constant 91. Careful, this may require changing the comparison operator."),
            lazy_problem().set_code("R6614").set_line(40)
            .set_text("Use \"(ord('Z') + 1)\" instead of using the magical constant 91."),
            lazy_problem().set_code("R6614").set_line(68)
            .set_text("Use \"ord('A')\" instead of using the magical constant 65."),
            lazy_problem().set_code("R6615").set_line(69)
            .set_text("Remove the call to 'ord' and compare to the string directly \"< 'A'\" instead of using "
                      "the magical constant 65. Careful, this may require changing the comparison operator."),
            lazy_problem().set_code("R6614").set_line(70)
            .set_text("Use \"ord('A')\" instead of using the magical constant 65.")
        ]),
        ("044262-vigenere.py", [  # TODO detect also 42 and 72
            lazy_problem().set_code("R6615").set_line(36)
            .set_text("Remove the call to 'ord' and compare to the string directly \"< 'A'\" instead of using "
                      "the magical constant 65. Careful, this may require changing the comparison operator."),
            lazy_problem().set_code("R6615").set_line(36)
            .set_text("Remove the call to 'ord' and compare to the string directly \"> '['\" instead of using "
                      "the magical constant 91. Careful, this may require changing the comparison operator."),
            lazy_problem().set_code("R6615").set_line(66),
            lazy_problem().set_code("R6615").set_line(66)
            .set_text("Remove the call to 'ord' and compare to the string directly \"> '['\" instead of using "
                      "the magical constant 91. Careful, this may require changing the comparison operator."),
        ]),
        ("048426-ipv4.py", [
            lazy_problem().set_code("R6614").set_line(4)
            .set_text("Use \"ord('0')\" instead of using the magical constant 48.")
        ]),
        ("hw34017.py", [
            lazy_problem().set_code("R6615").set_line(177),
            lazy_problem().set_code("R6615").set_line(177)
            .set_text("Remove the call to 'ord' and compare to the string directly \"< 'D'\" instead of using "
                      "the magical constant 68. Careful, this may require changing the comparison operator."),
            lazy_problem().set_code("R6615").set_line(177),
            lazy_problem().set_code("R6615").set_line(177)
            .set_text("Remove the call to 'ord' and compare to the string directly \"< '3'\" instead of using "
                      "the magical constant 51. Careful, this may require changing the comparison operator."),
            lazy_problem().set_code("R6614").set_line(178),
        ]),
    ])
    def test_ord_files(self, filename: str, expected_output: List[Problem]) -> None:
        apply_and_lint(
            filename,
            [
                Arg(Option.NO_FLAKE8, "on"),
                Arg(Option.PYLINT, "--disable=all"),
                Arg(Option.PYLINT, "--enable=use-ord-letter,use-literal-letter")
            ],
            expected_output
        )

    @pytest.mark.parametrize("filename,expected_output", [
        ("010666-prime.py", [
            lazy_problem().set_code("R6604").set_line(12).set_text("Do not use while loop with else."),
            lazy_problem().set_code("R6609").set_line(15).set_text("Use augmenting assignment: 'i += 1'"),
        ]),
        ("012889-pythagorean.py", [
            lazy_problem().set_code("R6602").set_line(23).set_text("Use integral division //."),
            lazy_problem().set_code("R6602").set_line(24),
            lazy_problem().set_code("R6602").set_line(25),
        ]),
        ("014290-coins.py", [
            lazy_problem().set_code("R6608").set_line(27).set_text("Redundant arithmetic: 1 * (value // 1)")
        ]),
        ("014494-next.py", [
            lazy_problem().set_code("R6602").set_line(5).set_text("Use integral division //."),
            lazy_problem().set_code("R6608").set_line(16).set_text("Redundant arithmetic: number**1"),
        ]),
        ("014661-p5_fibsum.py", [
            lazy_problem().set_code("R6609").set_line(15),
            lazy_problem().set_code("R6609").set_line(18),
            lazy_problem().set_code("R6609").set_line(19),
            lazy_problem().set_code("R6609").set_line(20),
            lazy_problem().set_code("R6611").set_line(23).set_text("Use else instead of elif."),
            lazy_problem().set_code("R6609").set_line(24),

        ]),
        ("017667-prime.py", [
            lazy_problem().set_code("R6604").set_line(6).set_text("Do not use for loop with else.")
        ]),
        ("022859-digit_sum.py", [
            lazy_problem().set_code("R6601").set_line(20)
            .set_text("Use lst.append(number % 7) instead of lst += [number % 7].")
        ]),
        ("024042-cards.py", [
            lazy_problem().set_code("R6609").set_line(9).set_text("Use augmenting assignment: 'number //= 10'"),
            lazy_problem().set_code("R6613").set_line(37)
            .set_text("Use 'is_valid_card(number)' directly rather than as 'is_valid_card(number) is True'."),
            lazy_problem().set_code("R6613").set_line(55),
        ]),
        ("024180-delete.py", [
            lazy_problem().set_code("R6606").set_line(7)
            .set_text("Remove the for loop, as it makes only one iteration.")
        ]),
        ("024233-cards.py", [
            lazy_problem().set_code("R6608").set_line(22).set_text("Redundant arithmetic: 0 + number"),
            lazy_problem().set_code("R6608").set_line(42).set_text("Redundant arithmetic: 0 + number")
        ]),
        ("024371-cards.py", [
            lazy_problem().set_code("R6606").set_line(12).set_text("Remove the for loop, as it makes no iterations.")
        ]),
        ("024371-workdays.py", [
            lazy_problem().set_code("R6606").set_line(29)
            .set_text("Remove the for loop, as it makes only one iteration."),
            lazy_problem().set_code("R6608").set_line(30).set_text("Redundant arithmetic: i + 0"),
        ]),
        ("024491-cards.py", [
            lazy_problem().set_code("R6606").set_line(86)
            .set_text("Remove the for loop, as it makes only one iteration.")
        ]),
        ("024657-bisection.py", [
            lazy_problem().set_code("R6608").set_line(6).set_text("Redundant arithmetic: 0 - eps"),
            lazy_problem().set_code("R6608").set_line(6).set_text("Redundant arithmetic: 0 + eps"),
            lazy_problem().set_code("R6608").set_line(8).set_text("Redundant arithmetic: 0 - eps"),
            lazy_problem().set_code("R6608").set_line(10).set_text("Redundant arithmetic: 0 + eps"),
            lazy_problem().set_code("R6608").set_line(12).set_text("Redundant arithmetic: 0 + eps"),
            lazy_problem().set_code("R6608").set_line(14).set_text("Redundant arithmetic: 0 + eps"),
        ]),
        ("032630-concat.py", [
            lazy_problem().set_code("R6608").set_line(5).set_text("Redundant arithmetic: [0] * 0")
        ]),
        ("041630-ipv4.py", [
            lazy_problem().set_code("R6603").set_line(15).set_text("Use isdecimal to test if string contains a number.")
        ]),
        ("044834-ipv4.py", [
            lazy_problem().set_code("R6603").set_line(15).set_text("Use isdecimal to test if string contains a number.")
        ]),
        ("064633-p4_histogram.py", [
            lazy_problem().set_code("R6611").set_line(35).set_text("Use else instead of elif.")
        ]),
        ("085286-p1_count.py", [
            lazy_problem().set_code("R6611").set_line(27).set_text("Use else instead of elif."),
            lazy_problem().set_code("R6612").set_line(30).set_text("Unreachable else."),
            lazy_problem().set_code("R6611").set_line(41).set_text("Use else instead of elif."),
            lazy_problem().set_code("R6612").set_line(44).set_text("Unreachable else."),
        ]),
        ("104174-ipv4_restore.py", [
            lazy_problem().set_code("R6610").set_line(38).set_text("Do not multiply list with mutable content.")
        ]),
        ("hw14358.py", [
            lazy_problem().set_code("R6609").set_line(29).set_text("Use augmenting assignment: 'result -= element'"),
            lazy_problem().set_code("R6609").set_line(31).set_text("Use augmenting assignment: 'b += 1'"),
            lazy_problem().set_code("R6609").set_line(34).set_text("Use augmenting assignment: 'a += 1'"),
            lazy_problem().set_code("R6608").set_line(44).set_text("Redundant arithmetic: num // num"),
            lazy_problem().set_code("R6609").set_line(50).set_text("Use augmenting assignment: 'count += 1'"),
            lazy_problem().set_code("R6609").set_line(51).set_text("Use augmenting assignment: 'number //= n'"),
            lazy_problem().set_code("R6609").set_line(53).set_text("Use augmenting assignment: 'primes += 1'"),
            lazy_problem().set_code("R6609").set_line(54),
            lazy_problem().set_code("R6609").set_line(56),
            lazy_problem().set_code("R6609").set_line(57),
        ]),
        ("hw34328.py", [
            lazy_problem().set_code("R6608").set_line(79).set_text("Redundant arithmetic: '' + str(count)")
        ]),
        ("hw34451.py", [
            lazy_problem().set_code("R6607").set_line(44)
            .set_text("Use exponentiation instead of repeated muliplication in i * i * i."),
            lazy_problem().set_code("R6607").set_line(47)
            .set_text("Use exponentiation instead of repeated muliplication in i * i * i."),
            lazy_problem().set_code("R6607").set_line(63)
            .set_text("Use exponentiation instead of repeated muliplication in i * i * i."),
            lazy_problem().set_code("R6607").set_line(66)
            .set_text("Use exponentiation instead of repeated muliplication in i * i * i."),
            lazy_problem().set_code("R6613").set_line(98)
            .set_text("Use 'not end' directly rather than as 'end is False'.")
        ]),
        ("hw35219.py", [
            lazy_problem().set_code("R6601").set_line(34)
            .set_text("Use coords_cw.append((-y, x)) instead of coords_cw += [(-y, x)]."),
            lazy_problem().set_code("R6601").set_line(41)
            .set_text("Use coords_ccw.append((y, -x)) instead of coords_ccw += [(y, -x)]."),
            lazy_problem().set_code("R6610").set_line(46).set_text("Do not multiply list with mutable content.")
        ]),
        ("hw45208.py", [
            lazy_problem().set_code("R6605").set_line(32),
            lazy_problem().set_code("R6611").set_line(45),
        ]),
        ("m2630.py", [
            lazy_problem().set_code("R6608").set_line(59).set_text("Redundant arithmetic: [0] * 0"),
            lazy_problem().set_code("R6608").set_line(79).set_text("Redundant arithmetic: num //= num"),
            lazy_problem().set_code("R6608").set_line(83).set_text("Redundant arithmetic: num //= num"),
            lazy_problem().set_code("R6608").set_line(86).set_text("Redundant arithmetic: num //= num"),
            lazy_problem().set_code("R6608").set_line(97).set_text("Redundant arithmetic: position * 1"),
            lazy_problem().set_code("R6613").set_line(144),
            lazy_problem().set_code("R6613").set_line(144),
        ]),
        ("m2650.py", [
            lazy_problem().set_code("R6609").set_line(11),
            lazy_problem().set_code("R6609").set_line(41),
            lazy_problem().set_code("R6609").set_line(44),
            lazy_problem().set_code("R6605").set_line(84).set_text("Use elif."),
        ]),
        ("m5435.py", [
            lazy_problem().set_code("R6609").set_line(7),
            lazy_problem().set_code("R6605").set_line(9).set_text("Use elif."),
            lazy_problem().set_code("R6608").set_line(53).set_text("Redundant arithmetic: valid += 0"),
            lazy_problem().set_code("R6608").set_line(64).set_text("Redundant arithmetic: valid += 0"),
        ]),
    ])
    def test_short_files(self, filename: str, expected_output: List[Problem]) -> None:
        apply_and_lint(
            filename,
            [
                Arg(Option.NO_FLAKE8, "on"),
                Arg(Option.PYLINT, "--disable=all"),
                Arg(Option.PYLINT, "--enable=short-problems")
            ],
            expected_output
        )



import pytest
from edulint.edulint import main
from os.path import join


def compare_output(monkeypatch, capsys, argv, output):
    monkeypatch.setattr("sys.argv", ["script"] + argv)
    main()
    captured = capsys.readouterr()
    import sys
    print(captured.err, file=sys.stderr)
    assert captured.out == output


@pytest.mark.parametrize("argv,output", [
    ([join("tests", "data", "hello_world.py")], ""),
    (
        [join("tests", "data", "custom_nonpep_assign.py")],
        "tests/data/custom_nonpep_assign.py:1:0: C0103 Constant name \"a\" doesn't conform to UPPER_CASE naming style\n"
        "tests/data/custom_nonpep_assign.py:1:2: E225 missing whitespace around operator\n"
    ),
    (
        [join("tests", "data", "custom_flake8_pylint.py")],
        "tests/data/custom_flake8_pylint.py:1:0: C0104 Disallowed name \"foo\"\n"
        "tests/data/custom_flake8_pylint.py:2:5: F841 local variable 'a' is assigned to but never used\n"
    ),
])
def test_single_file_stdout(monkeypatch, capsys, argv, output):
    compare_output(monkeypatch, capsys, argv, output)


@pytest.mark.parametrize("argv,output", [
    (
        [join("tests", "data", "hello_world.py"), join("tests", "data", "custom_nonpep_assign.py")],
        "****************** custom_nonpep_assign.py\n"
        "tests/data/custom_nonpep_assign.py:1:0: C0103 Constant name \"a\" doesn't conform to UPPER_CASE naming style\n"
        "tests/data/custom_nonpep_assign.py:1:2: E225 missing whitespace around operator\n"
    ),
    (
        [join("tests", "data", "custom_nonpep_assign.py"), join("tests", "data", "custom_flake8_pylint.py")],
        "****************** custom_nonpep_assign.py\n"
        "tests/data/custom_nonpep_assign.py:1:0: C0103 Constant name \"a\" doesn't conform to UPPER_CASE naming style\n"
        "tests/data/custom_nonpep_assign.py:1:2: E225 missing whitespace around operator\n"
        "****************** custom_flake8_pylint.py\n"
        "tests/data/custom_flake8_pylint.py:1:0: C0104 Disallowed name \"foo\"\n"
        "tests/data/custom_flake8_pylint.py:2:5: F841 local variable 'a' is assigned to but never used\n"
    ),
    (
        [join("tests", "data", "custom_flake8_pylint.py"), join("tests", "data", "custom_nonpep_assign.py")],
        "****************** custom_flake8_pylint.py\n"
        "tests/data/custom_flake8_pylint.py:1:0: C0104 Disallowed name \"foo\"\n"
        "tests/data/custom_flake8_pylint.py:2:5: F841 local variable 'a' is assigned to but never used\n"
        "****************** custom_nonpep_assign.py\n"
        "tests/data/custom_nonpep_assign.py:1:0: C0103 Constant name \"a\" doesn't conform to UPPER_CASE naming style\n"
        "tests/data/custom_nonpep_assign.py:1:2: E225 missing whitespace around operator\n"
    ),
])
def test_multiple_files_stdout(monkeypatch, capsys, argv, output):
    compare_output(monkeypatch, capsys, argv, output)


@pytest.mark.parametrize("argv,output", [
    (
        [join("tests", "data", "custom_nonpep_assign.py"), join("tests", "data", "custom_flake8_pylint_config.py")],
        "****************** custom_nonpep_assign.py\n"
        "tests/data/custom_nonpep_assign.py:1:0: C0103 Constant name \"a\" doesn't conform to UPPER_CASE naming style\n"
        "tests/data/custom_nonpep_assign.py:1:2: E225 missing whitespace around operator\n"
        "****************** custom_flake8_pylint_config.py\n"
        "tests/data/custom_flake8_pylint_config.py:1:0: C0114 Missing module docstring\n"
        "tests/data/custom_flake8_pylint_config.py:3:0: C0104 Disallowed name \"foo\"\n"
        "tests/data/custom_flake8_pylint_config.py:4:5: F841 local variable 'a' is assigned to but never used\n"
    ),
    (
        [join("tests", "data", "custom_flake8_pylint_config.py"), join("tests", "data", "custom_nonpep_assign.py")],
        "****************** custom_flake8_pylint_config.py\n"
        "tests/data/custom_flake8_pylint_config.py:1:0: C0114 Missing module docstring\n"
        "tests/data/custom_flake8_pylint_config.py:3:0: C0104 Disallowed name \"foo\"\n"
        "tests/data/custom_flake8_pylint_config.py:4:5: F841 local variable 'a' is assigned to but never used\n"
        "****************** custom_nonpep_assign.py\n"
        "tests/data/custom_nonpep_assign.py:1:0: C0103 Constant name \"a\" doesn't conform to UPPER_CASE naming style\n"
        "tests/data/custom_nonpep_assign.py:1:2: E225 missing whitespace around operator\n"
    ),
])
def test_different_configs(monkeypatch, capsys, argv, output):
    compare_output(monkeypatch, capsys, argv, output)


@pytest.mark.parametrize("argv,output", [
    (
        f"{join('tests', 'data', 'custom_flake8_pylint.py')} -o pylint=--enable=missing-module-docstring".split(),
        "tests/data/custom_flake8_pylint.py:1:0: C0114 Missing module docstring\n"
        "tests/data/custom_flake8_pylint.py:1:0: C0104 Disallowed name \"foo\"\n"
        "tests/data/custom_flake8_pylint.py:2:5: F841 local variable 'a' is assigned to but never used\n"
    ),
    (
        f"{join('tests', 'data', 'custom_flake8_pylint.py')} -o=pylint=--enable=missing-module-docstring".split(),
        "tests/data/custom_flake8_pylint.py:1:0: C0114 Missing module docstring\n"
        "tests/data/custom_flake8_pylint.py:1:0: C0104 Disallowed name \"foo\"\n"
        "tests/data/custom_flake8_pylint.py:2:5: F841 local variable 'a' is assigned to but never used\n"
    ),
    (
        f"{join('tests', 'data', 'custom_flake8_pylint.py')} --option pylint=--enable=missing-module-docstring".split(),
        "tests/data/custom_flake8_pylint.py:1:0: C0114 Missing module docstring\n"
        "tests/data/custom_flake8_pylint.py:1:0: C0104 Disallowed name \"foo\"\n"
        "tests/data/custom_flake8_pylint.py:2:5: F841 local variable 'a' is assigned to but never used\n"
    ),
    (
        f"{join('tests', 'data', 'custom_flake8_pylint.py')} --option=pylint=--enable=missing-module-docstring".split(),
        "tests/data/custom_flake8_pylint.py:1:0: C0114 Missing module docstring\n"
        "tests/data/custom_flake8_pylint.py:1:0: C0104 Disallowed name \"foo\"\n"
        "tests/data/custom_flake8_pylint.py:2:5: F841 local variable 'a' is assigned to but never used\n"
    ),
    (
        f"{join('tests', 'data', 'custom_flake8_pylint.py')} "
        "-o pylint=--enable=missing-module-docstring "
        "-o pylint=--disable=C0104".split(),
        "tests/data/custom_flake8_pylint.py:1:0: C0114 Missing module docstring\n"
        "tests/data/custom_flake8_pylint.py:2:5: F841 local variable 'a' is assigned to but never used\n"
    ),
    (
        [join('tests', 'data', 'custom_flake8_pylint.py'),
         "-o", "pylint=--enable=missing-module-docstring pylint=--disable=C0104"],
        "tests/data/custom_flake8_pylint.py:1:0: C0114 Missing module docstring\n"
        "tests/data/custom_flake8_pylint.py:2:5: F841 local variable 'a' is assigned to but never used\n"
    ),
])
def test_passing_cmd_args(monkeypatch, capsys, argv, output):
    compare_output(monkeypatch, capsys, argv, output)


@pytest.mark.parametrize("argv,output", [
    ([join("tests", "data", "custom_swap.py")], "")
])
def test_ib111_week(monkeypatch, capsys, argv, output):
    compare_output(monkeypatch, capsys, argv, output)



from typing import Dict
import os

import tomli

SCRIPT_PATH = os.path.dirname(os.path.abspath(__file__))
EXPLANATIONS_FILENAME = os.path.join(SCRIPT_PATH, 'explanations.toml')


def get_explanations() -> Dict[str, Dict[str, str]]:
    with open(EXPLANATIONS_FILENAME, "rb") as f:
        parsed_toml = tomli.load(f)
    return parsed_toml

# todo: add test to load explanantions (toml parser is sensitive to small mistakes)

if __name__ == "__main__":
    get_explanations()



import pytest
from edulint.linters import Linter
from edulint.options import Option
from edulint.config.arg import Arg
from edulint.config.config import Config
from edulint.linting.problem import Problem
from edulint.linting.linting import lint_one
from utils import lazy_problem, lazy_equal, get_tests_path, apply_and_lint, create_apply_and_lint
from typing import List


@pytest.mark.parametrize("filename,config,expected_output", [
    ("hello_world.py", Config(), []),
    ("z202817-zkouska.py", Config(), [
        lazy_problem().set_code("R6303").set_line(42),
        lazy_problem().set_code("R6205").set_line(82),
        lazy_problem().set_code("W0107").set_line(196)
    ]),
    ("z202817-zkouska.py", Config([Arg(Option.PYLINT, ["--enable=C0115"])]), [
        lazy_problem().set_code("R6303").set_line(42),
        lazy_problem().set_code("R6205").set_line(82),
        lazy_problem().set_code("C0115").set_line(122),
        lazy_problem().set_code("C0115").set_line(128),
        lazy_problem().set_code("W0107").set_line(196)
    ]),
    ("z202817-zkouska.py", Config([Arg(Option.PYLINT, ["--enable=C0115", "--disable=W0107"])]), [
        lazy_problem().set_code("R6303").set_line(42),
        lazy_problem().set_code("R6205").set_line(82),
        lazy_problem().set_code("C0115").set_line(122),
        lazy_problem().set_code("C0115").set_line(128)
    ]),
    ("z202817-zkouska.py",
     Config([Arg(Option.PYLINT, ["--disable=all"])]), []),
    ("002814-p1_trapezoid.py", Config(), [
        lazy_problem().set_code("F401").set_line(1),
        lazy_problem().set_code("F401").set_line(1),
        lazy_problem().set_code("E501").set_line(1),
        lazy_problem().set_code("W293").set_line(19),
        lazy_problem().set_code("E303").set_line(22),
    ])
])
def test_lint_basic(filename: str, config: Config, expected_output: List[Problem]) -> None:
    lazy_equal(lint_one(get_tests_path(filename), config), expected_output)


@pytest.mark.parametrize("filename,args,expected_output", [
    ("z202817-zkouska.py", [Arg(Option.ENHANCEMENT, "on")], [
        lazy_problem().set_code("R6609").set_line(10),
        lazy_problem().set_code("R6303").set_line(42),
        lazy_problem().set_code("R6208").set_line(76),
        lazy_problem().set_code("R6205").set_line(82),
        lazy_problem().set_code("W0107").set_line(196),
    ]),
    ("z202817-zkouska.py", [Arg(Option.PYTHON_SPECIFIC, "on")], [
        lazy_problem().set_code("R6303").set_line(42),
        lazy_problem().set_code("R6102").set_line(80),
        lazy_problem().set_code("R6205").set_line(82),
        lazy_problem().set_code("R6101").set_line(171),
        lazy_problem().set_code("C0123").set_line(172),
        lazy_problem().set_code("W0107").set_line(196),
    ]),
    ("z202817-zkouska.py", [Arg(Option.PYLINT, "--disable=all"), Arg(Option.PYTHON_SPECIFIC, "on")], [
        lazy_problem().set_code("R6102").set_line(80),
        lazy_problem().set_code("R6101").set_line(171),
        lazy_problem().set_code("C0123").set_line(172),
    ]),
    ("z202817-zkouska.py", [Arg(Option.PYLINT, "--disable=all"), Arg(Option.PYTHON_SPECIFIC, "off")], []),
    ("hw34406.py", [Arg(Option.PYLINT, "--disable=all"), Arg(Option.COMPLEXITY, "on")], [
        lazy_problem().set_line(240).set_code("R0913"),
        lazy_problem().set_line(266).set_code("R0911"),
        lazy_problem().set_line(266).set_code("R0912"),
        lazy_problem().set_line(343).set_code("R0912"),
        lazy_problem().set_line(343).set_code("R0915"),
        lazy_problem().set_line(387).set_code("R1702"),
        lazy_problem().set_line(387).set_code("R1702"),
        lazy_problem().set_line(387).set_code("R1702"),
    ])
])
def test_translations(filename: str, args: List[Arg], expected_output: List[Problem]) -> None:
    apply_and_lint(filename, args, expected_output)


class TestIB111Week:
    @pytest.mark.parametrize("lines,args,expected_output", [
        ([
            "def swap(a, b):",
            "    tmp = a",
            "    a = b",
            "    b = tmp",
        ], [Arg(Option.PYTHON_SPECIFIC, "on")], [
            lazy_problem().set_code("R1712").set_line(2),
        ]), ([
            "def swap(a, b):",
            "    tmp = a",
            "    a = b",
            "    b = tmp",
        ], [Arg(Option.PYTHON_SPECIFIC, "on"), Arg(Option.IB111_WEEK, "2")], [
        ]), ([
            "def swap(a, b):",
            "    tmp = a",
            "    a = b",
            "    b = tmp",
        ], [Arg(Option.PYTHON_SPECIFIC, "on"), Arg(Option.IB111_WEEK, "3")], [
            lazy_problem().set_code("R1712").set_line(2),
        ]), ([
            "def is_one_or_two(n):",
            "    if n == 1 or n == 2:",
            "        return True",
            "    return False"
        ], [Arg(Option.PYTHON_SPECIFIC, "on"), Arg(Option.IB111_WEEK, "6"), Arg(Option.PYLINT, "--disable=R6201")], [
        ]), ([
            "def is_one_or_two(n):",
            "    if n == 1 or n == 2:",
            "        return True",
            "    return False"
        ], [Arg(Option.PYTHON_SPECIFIC, "on"), Arg(Option.IB111_WEEK, "7"), Arg(Option.PYLINT, "--disable=R6201")], [
            lazy_problem().set_code("R1714").set_line(2),
        ])
    ])
    def test_ib111_week_custom(self, lines: List[str], args: List[Arg], expected_output: List[Problem]) -> None:
        create_apply_and_lint(lines, args, expected_output)


@pytest.mark.parametrize("filename,args,expected_output", [
    ("014186-p2_nested.py", [Arg(Option.PYTHON_SPECIFIC, "on")], [
        lazy_problem().set_code("C0103").set_line(20),
        lazy_problem().set_code("C0103").set_line(21),
        lazy_problem().set_code("C0103").set_line(27),
        lazy_problem().set_code("C0103").set_line(28),
        lazy_problem().set_code("C0103").set_line(31),
        lazy_problem().set_code("C0103").set_line(34),
        lazy_problem().set_code("W0622").set_line(48),
    ]),
    ("custom_pep_assign.py", [Arg(Option.PYTHON_SPECIFIC, "on")], [
        lazy_problem().set_code("C0103").set_line(1),
    ])
])
def test_invalid_name(filename: str, args: List[Arg], expected_output: List[Problem]) -> None:
    apply_and_lint(filename, args, expected_output)


@pytest.mark.parametrize("filename,args,expected_output", [
    ("014180-p5_fibsum.py", [Arg(Option.ALLOWED_ONECHAR_NAMES, "")], [
        lazy_problem().set_code("C0104").set_line(6),
        lazy_problem().set_code("C0104").set_line(14)
    ]),
    ("014180-p5_fibsum.py", [Arg(Option.ALLOWED_ONECHAR_NAMES, "n")], [
        lazy_problem().set_code("C0104").set_line(14)
    ]),
    ("014180-p5_fibsum.py", [Arg(Option.ALLOWED_ONECHAR_NAMES, "i")], [
        lazy_problem().set_code("C0104").set_line(6)
    ]),
    ("014180-p5_fibsum.py", [Arg(Option.ALLOWED_ONECHAR_NAMES, "in")], [
    ])
])
def test_allowed_onechar_names(filename: str, args: List[Arg], expected_output: List[Problem]) -> None:
    apply_and_lint(filename, args, expected_output)


@pytest.mark.parametrize("filename,args,expected_output", [
    ("105119-p5_template.py", [Arg(Option.PYLINT, "--disable=all"), Arg(Option.PYLINT, "--enable=R1714")], [
        lazy_problem().set_code("R1714").set_line(22)
        .set_text("Consider merging these comparisons with \"in\" to \"i not in '[]'\""),
        lazy_problem().set_code("R1714").set_line(35)
        .set_text("Consider merging these comparisons with \"in\" to 'i in (1, 2, 3)'"),
    ]),
])
def test_consider_using_in(filename: str, args: List[Arg], expected_output: List[Problem]) -> None:
    apply_and_lint(filename, args, expected_output)


class TestImproveFor:
    @pytest.mark.parametrize("lines,expected_output", [
        ([
            "A = list(range(10))",
            "B = []",
            "for x in range(len(A)):",
            "    for y in range(len(A)):",
            "        B.append(A[x])",
            "        B.append(y)"
        ], [
            lazy_problem().set_code("R6101").set_line(3)
            .set_text("Iterate directly: \"for var in A\" (with appropriate name for \"var\")"),
        ]), ([
            "A = list(range(10))",
            "for i in range(5, len(A)):",
            "    print(A[i])"
        ], [
        ]), ([
            "A = list(range(10))",
            "a = 5",
            "for i in range(a, len(A)):",
            "    print(A[i])"
        ], [
        ]), ([
            "A = list(range(10))",
            "a = 5",
            "for i in range(a, len(A)):",
            "    print(i, A[i])"
        ], [
        ]), ([
            "A = list(range(10))",
            "B = []",
            "for x in range(len(A)):",
            "    x += 1",
            "    B.append(A[x])",
            "    B.append(A[x + 1])"
        ], [
        ]), ([
            "A = list(range(10))",
            "B = []",
            "for x in range(len(A)):",
            "    B.append(A[x + 1])",
        ], [
        ]), ([
            "A = list(range(10))",
            "for x in range(len(A)):",
            "    A[x] = A[x] + 1"
        ], [
            lazy_problem().set_code("R6102").set_line(2)
            .set_text("Iterate using enumerate: \"for x, var in enumerate(A)\" (with appropriate name for \"var\")"),
        ]), ([
            "A = list(range(10))",
            "B = []",
            "for x in range(len(A)):",
            "    for x in range(len(B)):",
            "        B.append(A[x])"
        ], [
        ])
    ])
    def test_improve_for_custom(self, lines: List[str], expected_output: List[Problem]) -> None:
        create_apply_and_lint(
            lines,
            [Arg(Option.PYLINT, "--disable=all"), Arg(Option.PYLINT, "--enable=improve-for-loop")],
            expected_output
        )

    @pytest.mark.parametrize("filename,args,expected_output", [
        ("105119-p5_template.py", [Arg(Option.PYLINT, "--enable=iterate-directly")], [
        ]),
        ("015080-p4_geometry.py", [Arg(Option.PYLINT, "--enable=iterate-directly"),
                                   Arg(Option.PYLINT, "--disable=W0622,R1705,R1703,R6201,R6202")], [
        ]),
        ("014771-p2_nested.py", [Arg(Option.PYTHON_SPECIFIC, "on")], [
            lazy_problem().set_code("R6101").set_line(25)
            .set_text("Iterate directly: \"for var in A\" (with appropriate name for \"var\")"),
            lazy_problem().set_code("R6101").set_line(35)
            .set_text("Iterate directly: \"for var in A\" (with appropriate name for \"var\")"),
        ]),
        ("umime_count_a.py", [Arg(Option.PYLINT, "--enable=improve-for-loop"),
                              Arg(Option.FLAKE8, "--extend-ignore=E225")], [
            lazy_problem().set_code("R6101").set_line(3)
            .set_text("Iterate directly: \"for var in text\" (with appropriate name for \"var\")"),
        ]),
    ])
    def test_improve_for(self, filename: str, args: List[Arg], expected_output: List[Problem]) -> None:
        apply_and_lint(filename, [Arg(Option.PYLINT, "--disable=all")] + args, expected_output)


class TestPyTA:
    @pytest.mark.parametrize("lines,expected_output", [
        ([
            "from typing import Tuple",

            "Point = Tuple[int, int]"
        ], [
        ]),
        ([
            "if 0 == 1:",
            "    print('noooooooo')"
        ], [
            lazy_problem().set_line(1)
        ]),
    ])
    def test_forbidden_toplevel_code_custom(self, lines: List[str], expected_output: List[Problem]) -> None:
        create_apply_and_lint(
            lines,
            [Arg(Option.PYLINT, "--disable=all"),
             Arg(Option.PYLINT, "--enable=forbidden-top-level-code"),
             Arg(Option.NO_FLAKE8, "on")],
            [p.set_code("E9992") for p in expected_output]
        )


class TestNoGlobals:
    @pytest.mark.parametrize("lines,expected_output", [
        ([
            "x = 0",
            "def foo():",
            "    global x",
            "    x = 1"
        ], [
            lazy_problem().set_line(1)
            .set_text("Do not use global variables; you use x, modifying it for example at line 4.")
        ]),
        ([
            "x = 0",
            "def foo():",
            "    nonlocal x",
            "    x = 1"
        ], [
            lazy_problem().set_line(1)
            .set_text("Do not use global variables; you use x, modifying it for example at line 4.")
        ]),
        ([
            "x = 0",
            "def foo():",
            "    x = 1",
            "    def bar():",
            "        nonlocal x",
            "        x = 2"
        ], [
        ]),
        ([
            "x = []",
            "def foo():",
            "    x.append(1)"
        ], [
            lazy_problem().set_line(1)
            .set_text("Do not use global variables; you use x, modifying it for example at line 3.")
        ]),
        ([
            "x = None",
            "def foo():",
            "    x.y = 1"
        ], [
            lazy_problem().set_line(1)
            .set_text("Do not use global variables; you use x, modifying it for example at line 3.")
        ]),
        ([
            "if __name__ == '__main__':",
            "    test = 1",
            "    test = 2",
        ], [
        ]),
    ])
    def test_no_globals_custom(self, lines: List[str], expected_output: List[Problem]) -> None:
        create_apply_and_lint(
            lines,
            [Arg(Option.PYLINT, "--disable=all"),
             Arg(Option.PYLINT, "--enable=no-global-variables"),
             Arg(Option.NO_FLAKE8, "on")],
            [p.set_code("R6401") for p in expected_output]
        )

    @pytest.mark.parametrize("filename,expected_output", [
        ("054141-p1_database.py", []),
        ("054103-p1_database.py", []),
        ("hw14699.py", []),
        ("014466-fibfibsum.py", [
            lazy_problem().set_code("R6401").set_line(11)
            .set_text("Do not use global variables; you use fib, modifying it for example at line 18.")
        ]),
        ("021752-cards.py", [
            lazy_problem().set_code("R6401").set_line(7)
            .set_text("Do not use global variables; you use first_num, modifying it for example at line 27."),
            lazy_problem().set_code("R6401").set_line(8)
            .set_text("Do not use global variables; you use digits, modifying it for example at line 20."),
        ]),
        ("088952-p2_extremes.py", []),
        ("s10265-d_mancala.py", []),
    ])
    def test_no_globals_files(self, filename: str, expected_output: List[Problem]) -> None:
        apply_and_lint(
            filename,
            [Arg(Option.PYLINT, "--disable=all"), Arg(Option.PYLINT, "--enable=no-global-variables")],
            expected_output
        )


@pytest.mark.parametrize("filename,args,expected_output", [
    ("umime_count_a.py", [
        Arg(Option.PYTHON_SPECIFIC, "on"),
        Arg(Option.ALLOWED_ONECHAR_NAMES, ""),
        Arg(Option.ENHANCEMENT, "on"),
    ], [
        lazy_problem().set_code("C0104").set_line(2)
        .set_text("Disallowed single-character variable name \"a\", choose a more descriptive name"),
        lazy_problem().set_code("E225").set_line(2),
        lazy_problem().set_code("R6101").set_line(3)
        .set_text("Iterate directly: \"for var in text\" (with appropriate name for \"var\")"),
        lazy_problem().set_code("E225").set_line(4).set_column(19),
        lazy_problem().set_code("E225").set_line(4).set_column(35),
        lazy_problem().set_code("R6609").set_line(5)
        .set_text("Use augmenting assignment: 'a += 1'"),
    ]),
    ("umime_count_a_extended.py", [
        Arg(Option.PYTHON_SPECIFIC, "on"),
        Arg(Option.ALLOWED_ONECHAR_NAMES, ""),
        Arg(Option.ENHANCEMENT, "on"),
    ], [
        lazy_problem().set_code("R6201").set_line(2)
        .set_text("The if statement can be replaced with 'return ch == 'a' or ch == 'A''"),
        lazy_problem().set_code("R1714").set_line(2)
        .set_text("Consider merging these comparisons with \"in\" to \"ch in 'aA'\""),
        lazy_problem().set_code("C0104").set_line(9)
        .set_text("Disallowed single-character variable name \"a\", choose a more descriptive name"),
        lazy_problem().set_code("R6101").set_line(10)
        .set_text("Iterate directly: \"for var in text\" (with appropriate name for \"var\")"),
        lazy_problem().set_code("R6609").set_line(12)
        .set_text("Use augmenting assignment: 'a += 1'"),
    ])
])
def test_umime_count_a(filename: str, args: List[Arg], expected_output: List[Problem]) -> None:
    apply_and_lint(filename, args, expected_output)


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def fun():",
        "    return 42",
        "    pass"
    ], [
        lazy_problem().set_code("W0101")
    ]),
])
def test_overrides_custom(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(
        lines,
        [Arg(Option.NO_FLAKE8, "on")],
        expected_output
    )


@pytest.mark.filterwarnings("ignore:The 'default' argument to fields is deprecated. Use 'dump_default' instead.")
def test_problem_can_be_dumped_to_json() -> None:
    problem = Problem(source=Linter.FLAKE8, path='path', line=5, column=1, code='E303',
                      text='too many blank lines (3)', end_line=None, end_column=None)
    out = problem.to_json(indent=2, sort_keys=True)  # type: ignore
    assert out == """{
  "code": "E303",
  "column": 1,
  "end_column": null,
  "end_line": null,
  "line": 5,
  "path": "path",
  "source": "flake8",
  "symbol": null,
  "text": "too many blank lines (3)"
}"""

    out = Problem.schema().dumps([problem], indent=2, many=True, sort_keys=True)  # type: ignore
    assert out == """[
  {
    "code": "E303",
    "column": 1,
    "end_column": null,
    "end_line": null,
    "line": 5,
    "path": "path",
    "source": "flake8",
    "symbol": null,
    "text": "too many blank lines (3)"
  }
]"""



from astroid import nodes  # type: ignore
from typing import TYPE_CHECKING, List, Tuple, Iterator, Set

from pylint.checkers import BaseChecker  # type: ignore
from pylint.checkers.utils import only_required_for_messages

if TYPE_CHECKING:
    from pylint.lint import PyLinter  # type: ignore

from edulint.linting.checkers.utils import get_name, get_range_params, get_const_value
from edulint.linting.checkers.modified_listener import ModifiedListener


class ImproperLoop(BaseChecker):

    name = "improper-loop"
    msgs = {
        "R6301": (
            "The while condition can be replaced with '<negated %s>'",
            "no-while-true",
            "Emitted when the condition of a while loop is 'True' unnecessarily.",
        ),
        "R6302": (
            "Use tighter range boundaries, the %s iteration never happens.",
            "use-tighter-boundaries",
            "Emitted when the boundaries of a range can be made tighter."
        ),
        "R6303": (
            "Iterated structure %s is being modified inside the for loop body. Use while loop or iterate over a copy.",
            "modifying-iterated-structure",
            "Emitted when the structure that is being iterated over is being modified in the for loops body."
        ),
        "R6304": (
            "Changing the control variable %s of a for loop has no effect.",
            "changing-control-variable",
            "Emitted when the control variable of a for loop is being changed."
        ),
        "R6305": (
            "Use for loop.",
            "use-for-loop",
            "Emitted when a while loop can be transformed into a for loop."
        ),
        "R6306": (
            "Inner for loop shadows outer for loop's control variable %s.",
            "loop-shadows-control-variable",
            "Emitted when a for loop shadows control variable of an outer for loop."
        )
    }

    def _check_no_while_true(self, node: nodes.While) -> None:
        if not isinstance(node.test, nodes.Const) or not node.test.bool_value():
            return

        first = node.body[0]
        if isinstance(first, nodes.If) and isinstance(first.body[-1], nodes.Break):
            self.add_message("no-while-true", node=node, args=(first.test.as_string()))

    def _check_use_for_loop(self, node: nodes.While) -> None:
        def get_relevant_vals(node: nodes.NodeNG, result: Set[nodes.NodeNG] = None) -> Tuple[bool, Set[nodes.NodeNG]]:
            result = result if result is not None else set()

            if isinstance(node, nodes.Const):
                return True, result
            if isinstance(node, nodes.Name):
                result.add(node)
                return True, result
            if isinstance(node, nodes.BinOp):
                v1 = get_relevant_vals(node.left, result)
                v2 = get_relevant_vals(node.right, result)
                return v1 and v2, result
            if isinstance(node, nodes.Call) and node.func.as_string() == "len" and len(node.args) == 1:
                return get_relevant_vals(node.args[0], result)
            return False, result

        # TODO allow different increments?
        def adds_or_subtracts_one(node: nodes.NodeNG) -> bool:
            if not isinstance(node, nodes.AssignName):
                return False

            expr = node.parent
            if isinstance(expr, nodes.AugAssign):
                return expr.op in ("+=", "-=") and get_const_value(expr.value) == 1
            if isinstance(expr, nodes.Assign) and isinstance(expr.value, nodes.BinOp):
                binop = expr.value
                return binop.op in ("+", "-") \
                    and ((binop.left.as_string() == node.as_string() and get_const_value(binop.right) == 1)
                         or (binop.right.as_string() == node.as_string() and get_const_value(binop.left) == 1))
            return False

        test = node.test
        if not isinstance(test, nodes.Compare) or len(test.ops) != 1 or test.ops[0][0] not in ("<", "<=", ">", ">="):
            return

        lt, rt = test.left, test.ops[0][1]
        lt_decomposable, lt_vals = get_relevant_vals(lt)
        rt_decomposable, rt_vals = get_relevant_vals(rt)
        if not lt_decomposable or not rt_decomposable:
            return

        all_vals = lt_vals | rt_vals
        listener = ModifiedListener(all_vals)
        listener.visit_many(node.body)

        all_modifiers = [(val in lt_vals, listener.get_all_modifiers(val)) for val in all_vals]
        nonempty_modifiers = [(from_lt, modifiers) for (from_lt, modifiers) in all_modifiers if len(modifiers) > 0]

        if len(nonempty_modifiers) != 1:
            return

        from_lt, only_modifiers = nonempty_modifiers[0]
        if len(only_modifiers) != 1:
            return

        only_modifier = only_modifiers[0]

        if self._get_block_line(only_modifier).parent != node or not adds_or_subtracts_one(only_modifier) \
                or (only_modifier.as_string() != lt.as_string() and only_modifier.as_string() != rt.as_string()):
            return

        self.add_message("use-for-loop", node=node)

    @only_required_for_messages("no-while-true", "use-for-loop")
    def visit_while(self, node: nodes.While) -> None:
        self._check_no_while_true(node)
        self._check_use_for_loop(node)

    def _check_use_tighter_bounds(self, node: nodes.For) -> None:
        def compares_equality(node: nodes.NodeNG) -> bool:
            return isinstance(node, nodes.Compare) and len(node.ops) == 1 and node.ops[0][0] == "=="

        def compares_to_start(node: nodes.Compare, var: str, start: nodes.NodeNG) -> bool:
            left, (_, right) = node.left, node.ops[0]
            left, right = left.as_string(), right.as_string()
            return (left == var and right == start.as_string()) or (left == start.as_string() and right == var)

        def is_last_before_end(node: nodes.NodeNG, stop: nodes.NodeNG, step: nodes.NodeNG) -> bool:
            const_node = get_const_value(node)
            const_stop = get_const_value(stop)
            const_step = get_const_value(step)
            return (isinstance(node, nodes.BinOp) and node.op == "-"
                    and node.left.as_string() == stop.as_string() and node.right.as_string() == step.as_string()) \
                or (isinstance(const_node, int) and isinstance(const_stop, int) and isinstance(const_step, int)
                    and const_node == const_stop - const_step)

        def compares_to_last_before_end(node: nodes.Compare, var: str, stop: nodes.NodeNG, step: nodes.NodeNG) -> bool:
            left, (_, right) = node.left, node.ops[0]
            return (left.as_string() == var and is_last_before_end(right, stop, step)) \
                or (is_last_before_end(left, stop, step) and right.as_string() == var)

        def relevant_nodes(body: List[nodes.NodeNG]) -> Iterator[nodes.If]:
            first = body[0]
            if isinstance(first, nodes.If):
                yield first
                while first.has_elif_block():
                    first = first.orelse[0]
                    yield first

            if len(body) > 1 and isinstance(body[-1], nodes.If):
                yield body[-1]

        range_params = get_range_params(node.iter)
        if range_params is None:
            return

        start, stop, step = range_params

        var = node.target.as_string()
        if isinstance(node.body[0], nodes.If):
            if_ = node.body[0]
            test = if_.test

            if compares_equality(test) and len(if_.body) == 1 and type(if_.body[-1]) in (nodes.Break, nodes.Continue):
                if compares_to_start(test, var, start):
                    self.add_message("use-tighter-boundaries", node=node, args=("first",))
                elif compares_to_last_before_end(test, var, stop, step):
                    self.add_message("use-tighter-boundaries", node=node, args=("last",))
            # TODO expand, tips: len(node.body) == 1, isinstance(if_.body[-1], nodes.Return), allow nodes in the middle,
            #                    allow longer bodies

    @staticmethod
    def _get_block_line(node: nodes.NodeNG):
        while not isinstance(node, nodes.Statement):
            node = node.parent
        return node

    @staticmethod
    def _get_last_block_line(node: nodes.NodeNG):
        node = ImproperLoop._get_block_line(node)

        while node.next_sibling() is not None:
            node = node.next_sibling()
        return node

    def _check_modifying_iterable(self, node: nodes.For) -> None:
        iterated = node.iter
        if type(iterated) not in (nodes.Name, nodes.Attribute):  # TODO allow any node type
            return

        listener = ModifiedListener({iterated})
        listener.visit_many(node.body)

        for modifier in listener.get_sure_modifiers(iterated):
            if isinstance(modifier, nodes.Call) \
                    and type(self._get_last_block_line(modifier)) not in (nodes.Break, nodes.Return):
                self.add_message("modifying-iterated-structure", node=modifier, args=(get_name(iterated),))

    def _check_control_variable_changes(self, node: nodes.For) -> None:
        def is_last_block(node: nodes.NodeNG, for_: nodes.For) -> bool:
            stmt = self._get_block_line(node)
            while not isinstance(node, nodes.Module):
                if type(stmt) in (nodes.For, nodes.While):
                    return False
                if stmt == for_.body[-1]:
                    return True
                last_block_stmt = self._get_last_block_line(stmt)
                if stmt != last_block_stmt:
                    return False
                stmt = last_block_stmt.parent
            return False

        range_params = get_range_params(node.iter)
        if range_params is None:
            return

        control_var = node.target
        if control_var.as_string().startswith("_"):
            return

        listener = ModifiedListener({control_var})
        listener.visit_many(node.body)

        for modifier in listener.get_all_modifiers(control_var):
            mod_statement = self._get_block_line(modifier)
            if isinstance(mod_statement, nodes.For) and mod_statement.target.as_string() == control_var.as_string():
                self.add_message("loop-shadows-control-variable", node=mod_statement, args=(modifier.as_string()))
            if is_last_block(mod_statement, node):
                self.add_message("changing-control-variable", node=mod_statement, args=(control_var.as_string(),))

    @only_required_for_messages("use-tighter-boundaries", "modifying-iterated-structure", "changing-control-variable",
                                "loop-shadows-control-variable")
    def visit_for(self, node: nodes.For) -> None:
        self._check_use_tighter_bounds(node)
        self._check_modifying_iterable(node)
        self._check_control_variable_changes(node)


def register(linter: "PyLinter") -> None:
    """This required method auto registers the checker during initialization.
    :param linter: The linter to register the checker to.
    """
    linter.register_checker(ImproperLoop(linter))
